@@ -21,7 +21,11 @@
 
 #ifndef lint
 static const char rcsid[] =
-    "@(#) $Header: print-ip.c,v 1.66 97/05/28 12:51:43 leres Exp $ (LBL)";
+    "@(#) $Header: /tcpdump/master/tcpdump/print-ip.c,v 1.92 2001/01/02 23:00:01 guy Exp $ (LBL)";
+#endif
+
+#ifdef HAVE_CONFIG_H
+#include "config.h"
 #endif
 
 #include <sys/param.h>
@@ -29,17 +33,7 @@ static const char rcsid[] =
 #include <sys/socket.h>
 
 #include <netinet/in.h>
-#include <netinet/in_systm.h>
-#include <netinet/ip.h>
-#include <netinet/ip_var.h>
-#include <netinet/udp.h>
-#include <netinet/udp_var.h>
-#include <netinet/tcp.h>
-#include <netinet/tcpip.h>
 
-#ifdef HAVE_MALLOC_H
-#include <malloc.h>
-#endif
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -49,178 +43,13 @@ static const char rcsid[] =
 #include "interface.h"
 #include "extract.h"			/* must come after interface.h */
 
+#include "ip.h"
+
 /* Compatibility */
 #ifndef	IPPROTO_ND
 #define	IPPROTO_ND	77
 #endif
 
-#ifndef IN_CLASSD
-#define IN_CLASSD(i) (((int32_t)(i) & 0xf0000000) == 0xe0000000)
-#endif
-
-/* (following from ipmulti/mrouted/prune.h) */
-
-/*
- * The packet format for a traceroute request.
- */
-struct tr_query {
-	u_int  tr_src;			/* traceroute source */
-	u_int  tr_dst;			/* traceroute destination */
-	u_int  tr_raddr;		/* traceroute response address */
-#ifdef WORDS_BIGENDIAN
-	struct {
-		u_int   ttl : 8;	/* traceroute response ttl */
-		u_int   qid : 24;	/* traceroute query id */
-	} q;
-#else
-	struct {
-		u_int	qid : 24;	/* traceroute query id */
-		u_int	ttl : 8;	/* traceroute response ttl */
-	} q;
-#endif
-};
-
-#define tr_rttl q.ttl
-#define tr_qid  q.qid
-
-/*
- * Traceroute response format.  A traceroute response has a tr_query at the
- * beginning, followed by one tr_resp for each hop taken.
- */
-struct tr_resp {
-	u_int tr_qarr;			/* query arrival time */
-	u_int tr_inaddr;		/* incoming interface address */
-	u_int tr_outaddr;		/* outgoing interface address */
-	u_int tr_rmtaddr;		/* parent address in source tree */
-	u_int tr_vifin;			/* input packet count on interface */
-	u_int tr_vifout;		/* output packet count on interface */
-	u_int tr_pktcnt;		/* total incoming packets for src-grp */
-	u_char  tr_rproto;		/* routing proto deployed on router */
-	u_char  tr_fttl;		/* ttl required to forward on outvif */
-	u_char  tr_smask;		/* subnet mask for src addr */
-	u_char  tr_rflags;		/* forwarding error codes */
-};
-
-/* defs within mtrace */
-#define TR_QUERY 1
-#define TR_RESP	2
-
-/* fields for tr_rflags (forwarding error codes) */
-#define TR_NO_ERR	0
-#define TR_WRONG_IF	1
-#define TR_PRUNED	2
-#define TR_OPRUNED	3
-#define TR_SCOPED	4
-#define TR_NO_RTE	5
-#define TR_NO_FWD	7
-#define TR_NO_SPACE	0x81
-#define TR_OLD_ROUTER	0x82
-
-/* fields for tr_rproto (routing protocol) */
-#define TR_PROTO_DVMRP	1
-#define TR_PROTO_MOSPF	2
-#define TR_PROTO_PIM	3
-#define TR_PROTO_CBT	4
-
-static void print_mtrace(register const u_char *bp, register u_int len)
-{
-	register struct tr_query *tr = (struct tr_query *)(bp + 8);
-
-	printf("mtrace %d: %s to %s reply-to %s", tr->tr_qid,
-		ipaddr_string(&tr->tr_src), ipaddr_string(&tr->tr_dst),
-		ipaddr_string(&tr->tr_raddr));
-	if (IN_CLASSD(ntohl(tr->tr_raddr)))
-		printf(" with-ttl %d", tr->tr_rttl);
-}
-
-static void print_mresp(register const u_char *bp, register u_int len)
-{
-	register struct tr_query *tr = (struct tr_query *)(bp + 8);
-
-	printf("mresp %d: %s to %s reply-to %s", tr->tr_qid,
-		ipaddr_string(&tr->tr_src), ipaddr_string(&tr->tr_dst),
-		ipaddr_string(&tr->tr_raddr));
-	if (IN_CLASSD(ntohl(tr->tr_raddr)))
-		printf(" with-ttl %d", tr->tr_rttl);
-}
-
-static void
-igmp_print(register const u_char *bp, register u_int len,
-	   register const u_char *bp2)
-{
-	register const struct ip *ip;
-
-	ip = (const struct ip *)bp2;
-        (void)printf("%s > %s: ",
-		ipaddr_string(&ip->ip_src),
-		ipaddr_string(&ip->ip_dst));
-
-	TCHECK2(bp[0], 8);
-	switch (bp[0]) {
-	case 0x11:
-		(void)printf("igmp query");
-		if (*(int *)&bp[4])
-			(void)printf(" [gaddr %s]", ipaddr_string(&bp[4]));
-		if (len != 8)
-			(void)printf(" [len %d]", len);
-		break;
-	case 0x12:
-		(void)printf("igmp report %s", ipaddr_string(&bp[4]));
-		if (len != 8)
-			(void)printf(" [len %d]", len);
-		break;
-	case 0x16:
-		(void)printf("igmp nreport %s", ipaddr_string(&bp[4]));
-		break;
-	case 0x17:
-		(void)printf("igmp leave %s", ipaddr_string(&bp[4]));
-		break;
-	case 0x13:
-		(void)printf("igmp dvmrp");
-		if (len < 8)
-			(void)printf(" [len %d]", len);
-		else
-			dvmrp_print(bp, len);
-		break;
-	case 0x14:
-		(void)printf("igmp pim");
-		pim_print(bp, len);
-  		break;
-	case 0x1e:
-		print_mresp(bp, len);
-		break;
-	case 0x1f:
-		print_mtrace(bp, len);
-		break;
-	default:
-		(void)printf("igmp-%d", bp[0] & 0xf);
-		break;
-	}
-	if ((bp[0] >> 4) != 1)
-		(void)printf(" [v%d]", bp[0] >> 4);
-
-	TCHECK2(bp[0], len);
-	if (vflag) {
-		/* Check the IGMP checksum */
-		u_int32_t sum = 0;
-		int count;
-		const u_short *sp = (u_short *)bp;
-		
-		for (count = len / 2; --count >= 0; )
-			sum += *sp++;
-		if (len & 1)
-			sum += ntohs(*(u_char *) sp << 8);
-		while (sum >> 16)
-			sum = (sum & 0xffff) + (sum >> 16);
-		sum = 0xffff & ~sum;
-		if (sum != 0)
-			printf(" bad igmp cksum %x!", EXTRACT_16BITS(&bp[2]));
-	}
-	return;
-trunc:
-	fputs("[|igmp]", stdout);
-}
-
 /*
  * print the recorded route in an IP RR, LSRR or SSRR option.
  */
@@ -246,6 +75,62 @@ ip_printroute(const char *type, register const u_char *cp, u_int length)
 	printf("%s}", ptr == len? "#" : "");
 }
 
+static void
+ip_printts(register const u_char *cp, u_int length)
+{
+	register u_int ptr = cp[2] - 1;
+	register u_int len = 0;
+	int hoplen;
+	char *type;
+
+	printf(" TS{");
+	hoplen = ((cp[3]&0xF) != IPOPT_TS_TSONLY) ? 8 : 4;
+	if ((length - 4) & (hoplen-1))
+		printf("[bad length %d]", length);
+	if (ptr < 4 || ((ptr - 4) & (hoplen-1)) || ptr > length + 1)
+		printf("[bad ptr %d]", cp[2]);
+	switch (cp[3]&0xF) {
+	case IPOPT_TS_TSONLY:
+		printf("TSONLY");
+		break;
+	case IPOPT_TS_TSANDADDR:
+		printf("TS+ADDR");
+		break;
+	/*
+	 * prespecified should really be 3, but some ones might send 2
+	 * instead, and the IPOPT_TS_PRESPEC constant can apparently
+	 * have both values, so we have to hard-code it here.
+	 */
+
+	case 2:
+		printf("PRESPEC2.0");
+		break;
+	case 3:			/* IPOPT_TS_PRESPEC */
+		printf("PRESPEC");
+		break;
+	default:	
+		printf("[bad ts type %d]", cp[3]&0xF);
+		goto done;
+	}
+
+	type = " ";
+	for (len = 4; len < length; len += hoplen) {
+		if (ptr == len)
+			type = " ^ ";
+		printf("%s%d@%s", type, EXTRACT_32BITS(&cp[len+hoplen-4]),
+		       hoplen!=8 ? "" : ipaddr_string(&cp[len]));
+		type = " ";
+	}
+
+done:
+	printf("%s", ptr == len ? " ^ " : "");
+
+	if (cp[3]>>4)
+		printf(" [%d hops not recorded]} ", cp[3]>>4);
+	else
+		printf("}");
+}
+
 /*
  * print IP options.
  */
@@ -257,7 +142,15 @@ ip_optprint(register const u_char *cp, u_int length)
 	for (; length > 0; cp += len, length -= len) {
 		int tt = *cp;
 
-		len = (tt == IPOPT_NOP || tt == IPOPT_EOL) ? 1 : cp[1];
+		if (tt == IPOPT_NOP || tt == IPOPT_EOL)
+			len = 1;
+		else {
+			if (&cp[1] >= snapend) {
+				printf("[|ip]");
+				return;
+			}
+			len = cp[1];
+		}
 		if (len <= 0) {
 			printf("[|ip op len %d]", len);
 			return;
@@ -279,15 +172,17 @@ ip_optprint(register const u_char *cp, u_int length)
 			break;
 
 		case IPOPT_TS:
-			printf(" TS{%d}", len);
+			ip_printts(cp, len);
 			break;
 
+#ifndef IPOPT_SECURITY
+#define IPOPT_SECURITY 130
+#endif /* IPOPT_SECURITY */
 		case IPOPT_SECURITY:
 			printf(" SECURITY{%d}", len);
 			break;
 
 		case IPOPT_RR:
-			printf(" RR{%d}=", len);
 			ip_printroute("RR", cp, len);
 			break;
 
@@ -299,6 +194,17 @@ ip_optprint(register const u_char *cp, u_int length)
 			ip_printroute("LSRR", cp, len);
 			break;
 
+#ifndef IPOPT_RA
+#define IPOPT_RA 148		/* router alert */
+#endif
+		case IPOPT_RA:
+			printf(" RA");
+			if (len != 4)
+				printf("{%d}", len);
+			else if (cp[2] || cp[3])
+				printf("%d.%d", cp[2], cp[3]);
+ 			break;
+
 		default:
 			printf(" IPOPT-%d{%d}", cp[0], len);
 			break;
@@ -310,23 +216,34 @@ ip_optprint(register const u_char *cp, u_int length)
  * compute an IP header checksum.
  * don't modifiy the packet.
  */
-static int
-in_cksum(const struct ip *ip)
+u_short
+in_cksum(const u_short *addr, register int len, u_short csum)
 {
-	register const u_short *sp = (u_short *)ip;
-	register u_int32_t sum = 0;
-	register int count;
+	int nleft = len;
+	const u_short *w = addr;
+	u_short answer;
+	int sum = csum;
+
+ 	/*
+	 *  Our algorithm is simple, using a 32 bit accumulator (sum),
+	 *  we add sequential 16 bit words to it, and at the end, fold
+	 *  back all the carry bits from the top 16 bits into the lower
+	 *  16 bits.
+ 	 */
+	while (nleft > 1)  {
+		sum += *w++;
+		nleft -= 2;
+	}
+	if (nleft == 1)
+		sum += htons(*(u_char *)w<<8);
 
 	/*
-	 * No need for endian conversions.
+	 * add back carry outs from top 16 bits to low 16 bits
 	 */
-	for (count = ip->ip_hl * 2; --count >= 0; )
-		sum += *sp++;
-	while (sum > 0xffff)
-		sum = (sum & 0xffff) + (sum >> 16);
-	sum = ~sum & 0xffff;
-
-	return (sum);
+	sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
+	sum += (sum >> 16);			/* add carry */
+	answer = ~sum;				/* truncate to 16 bits */
+	return (answer);
 }
 
 /*
@@ -336,8 +253,10 @@ void
 ip_print(register const u_char *bp, register u_int length)
 {
 	register const struct ip *ip;
-	register u_int hlen, len, off;
+	register u_int hlen, len, len0, off;
 	register const u_char *cp;
+	u_char nh;
+	int advance;
 
 	ip = (const struct ip *)bp;
 #ifdef LBL_ALIGN
@@ -374,13 +293,18 @@ ip_print(register const u_char *bp, register u_int length)
 		(void)printf("truncated-ip %d", length);
 		return;
 	}
-	hlen = ip->ip_hl * 4;
+	hlen = IP_HL(ip) * 4;
+	if (hlen < sizeof (struct ip)) {
+		(void)printf("bad-hlen %d", hlen);
+		return;
+	}
 
 	len = ntohs(ip->ip_len);
 	if (length < len)
 		(void)printf("truncated-ip - %d bytes missing!",
 			len - length);
 	len -= hlen;
+	len0 = len;
 
 	/*
 	 * If this is fragment zero, hand it to the next higher
@@ -389,18 +313,65 @@ ip_print(register const u_char *bp, register u_int length)
 	off = ntohs(ip->ip_off);
 	if ((off & 0x1fff) == 0) {
 		cp = (const u_char *)ip + hlen;
-		switch (ip->ip_p) {
+		nh = ip->ip_p;
+
+		if (nh != IPPROTO_TCP && nh != IPPROTO_UDP) {
+			(void)printf("%s > %s: ", ipaddr_string(&ip->ip_src),
+				ipaddr_string(&ip->ip_dst));
+		}
+again:
+		switch (nh) {
+
+#ifndef IPPROTO_AH
+#define IPPROTO_AH	51
+#endif
+		case IPPROTO_AH:
+			nh = *cp;
+			advance = ah_print(cp, (const u_char *)ip);
+			cp += advance;
+			len -= advance;
+			goto again;
+
+#ifndef IPPROTO_ESP
+#define IPPROTO_ESP	50
+#endif
+		case IPPROTO_ESP:
+		    {
+			int enh;
+			advance = esp_print(cp, (const u_char *)ip, &enh);
+			cp += advance;
+			len -= advance;
+			if (enh < 0)
+				break;
+			nh = enh & 0xff;
+			goto again;
+		    }
+
+#ifndef IPPROTO_IPCOMP
+#define IPPROTO_IPCOMP	108
+#endif
+		case IPPROTO_IPCOMP:
+		    {
+			int enh;
+			advance = ipcomp_print(cp, (const u_char *)ip, &enh);
+			cp += advance;
+			len -= advance;
+			if (enh < 0)
+				break;
+			nh = enh & 0xff;
+			goto again;
+		    }
 
 		case IPPROTO_TCP:
-			tcp_print(cp, len, (const u_char *)ip);
+			tcp_print(cp, len, (const u_char *)ip, (off &~ 0x6000));
 			break;
 
 		case IPPROTO_UDP:
-			udp_print(cp, len, (const u_char *)ip);
+			udp_print(cp, len, (const u_char *)ip, (off &~ 0x6000));
 			break;
 
 		case IPPROTO_ICMP:
-			icmp_print(cp, (const u_char *)ip);
+			icmp_print(cp, len, (const u_char *)ip);
 			break;
 
 #ifndef IPPROTO_IGRP
@@ -411,8 +382,10 @@ ip_print(register const u_char *bp, register u_int length)
 			break;
 
 		case IPPROTO_ND:
+#if 0
 			(void)printf("%s > %s:", ipaddr_string(&ip->ip_src),
 				ipaddr_string(&ip->ip_dst));
+#endif
 			(void)printf(" nd %d", len);
 			break;
 
@@ -436,10 +409,12 @@ ip_print(register const u_char *bp, register u_int length)
 
 		case 4:
 			/* DVMRP multicast tunnel (ip-in-ip encapsulation) */
+#if 0
 			if (vflag)
 				(void)printf("%s > %s: ",
 					     ipaddr_string(&ip->ip_src),
 					     ipaddr_string(&ip->ip_dst));
+#endif
 			ip_print(cp, len);
 			if (! vflag) {
 				printf(" (ipip)");
@@ -447,6 +422,27 @@ ip_print(register const u_char *bp, register u_int length)
 			}
 			break;
 
+#ifdef INET6
+#ifndef IP6PROTO_ENCAP
+#define IP6PROTO_ENCAP 41
+#endif
+		case IP6PROTO_ENCAP:
+			/* ip6-in-ip encapsulation */
+#if 0
+			if (vflag)
+				(void)printf("%s > %s: ",
+					     ipaddr_string(&ip->ip_src),
+					     ipaddr_string(&ip->ip_dst));
+#endif
+			ip6_print(cp, len);
+			if (! vflag) {
+				printf(" (encap)");
+				return;
+			}
+			break;
+#endif /*INET6*/
+
+
 #ifndef IPPROTO_GRE
 #define IPPROTO_GRE 47
 #endif
@@ -463,18 +459,60 @@ ip_print(register const u_char *bp, register u_int length)
   			}
   			break;
 
+#ifndef IPPROTO_MOBILE
+#define IPPROTO_MOBILE 55
+#endif
+		case IPPROTO_MOBILE:
+			if (vflag)
+				(void)printf("mobile %s > %s: ",
+					     ipaddr_string(&ip->ip_src),
+					     ipaddr_string(&ip->ip_dst));
+			mobile_print(cp, len);
+			if (! vflag) {
+				printf(" (mobile encap)");
+				return;
+			}
+			break;
+
+#ifndef IPPROTO_PIM
+#define IPPROTO_PIM	103
+#endif
+		case IPPROTO_PIM:
+			pim_print(cp, len);
+			break;
+
+#ifndef IPPROTO_VRRP
+#define IPPROTO_VRRP	112
+#endif
+		case IPPROTO_VRRP:
+			if (vflag)
+				(void)printf("vrrp %s > %s: ",
+					     ipaddr_string(&ip->ip_src),
+					     ipaddr_string(&ip->ip_dst));
+			vrrp_print(cp, len, ip->ip_ttl);
+			break;
+
 		default:
+#if 0
 			(void)printf("%s > %s:", ipaddr_string(&ip->ip_src),
 				ipaddr_string(&ip->ip_dst));
-			(void)printf(" ip-proto-%d %d", ip->ip_p, len);
+#endif
+			(void)printf(" ip-proto-%d %d", nh, len);
 			break;
 		}
 	}
+
+ 	/* Ultra quiet now means that all this stuff should be suppressed */
+ 	/* res 3-Nov-98 */
+ 	if (qflag > 1) return;
+
+
 	/*
 	 * for fragmented datagrams, print id:size@offset.  On all
 	 * but the last stick a "+".  For unfragmented datagrams, note
 	 * the don't fragment flag.
 	 */
+	len = len0;	/* get the original length */
 	if (off & 0x3fff) {
 		/*
 		 * if this isn't the first frag, we're missing the
@@ -483,14 +521,30 @@ ip_print(register const u_char *bp, register u_int length)
 		if (off & 0x1fff)
 			(void)printf("%s > %s:", ipaddr_string(&ip->ip_src),
 				      ipaddr_string(&ip->ip_dst));
-		(void)printf(" (frag %d:%d@%d%s)", ntohs(ip->ip_id), len,
+#ifndef IP_MF
+#define IP_MF 0x2000
+#endif /* IP_MF */
+#ifndef IP_DF
+#define IP_DF 0x4000
+#endif /* IP_DF */
+		(void)printf(" (frag %d:%u@%d%s)", ntohs(ip->ip_id), len,
 			(off & 0x1fff) * 8,
 			(off & IP_MF)? "+" : "");
+
 	} else if (off & IP_DF)
 		(void)printf(" (DF)");
 
-	if (ip->ip_tos)
-		(void)printf(" [tos 0x%x]", (int)ip->ip_tos);
+	if (ip->ip_tos) {
+		(void)printf(" [tos 0x%x", (int)ip->ip_tos);
+		/* ECN bits */
+		if (ip->ip_tos&0x02) {
+			(void)printf(",ECT");
+			if (ip->ip_tos&0x01)
+				(void)printf(",CE");
+		}
+		(void)printf("] ");
+	}
+
 	if (ip->ip_ttl <= 1)
 		(void)printf(" [ttl %d]", (int)ip->ip_ttl);
 
@@ -507,8 +561,10 @@ ip_print(register const u_char *bp, register u_int length)
 			(void)printf("%sid %d", sep, (int)ntohs(ip->ip_id));
 			sep = ", ";
 		}
+		(void)printf("%slen %d", sep, (int)ntohs(ip->ip_len));
+		sep = ", ";
 		if ((u_char *)ip + hlen <= snapend) {
-			sum = in_cksum(ip);
+			sum = in_cksum((const u_short *)ip, hlen, 0);
 			if (sum != 0) {
 				(void)printf("%sbad cksum %x!", sep,
 					     ntohs(ip->ip_sum));
@@ -522,3 +578,29 @@ ip_print(register const u_char *bp, register u_int length)
 		printf(")");
 	}
 }
+
+void
+ipN_print(register const u_char *bp, register u_int length)
+{
+	struct ip *ip, hdr;
+
+	ip = (struct ip *)bp;
+	if (length < 4) {
+		(void)printf("truncated-ip %d", length);
+		return;
+	}
+	memcpy (&hdr, (char *)ip, 4);
+	switch (IP_V(&hdr)) {
+	case 4:
+	    ip_print (bp, length);
+	    return;
+#ifdef INET6
+	case 6:
+	    ip6_print (bp, length);
+	    return;
+#endif
+	default:
+	    (void)printf("unknown ip %d", IP_V(&hdr));
+	    return;
+	}
+}
